import magma as m

def Enum(**kwargs):
    max_value = max(value for value in kwargs.values())
    num_bits = max(max_value.bit_length(), 1)
    type_ = m.Bits(num_bits)
    for key, value in kwargs.items():
        setattr(type_, key, m.bits(value, num_bits))
    return type_



from enum import Enum, auto
from .is_definition import isdefinition
from .logging import warning, error
from .passes import DefinitionPass


class MultipleDefinitionException(Exception):
    pass


class UniquificationMode(Enum):
    WARN = auto()
    ERROR = auto()
    UNIQUIFY = auto()


class UniquificationPass(DefinitionPass):
    def __init__(self, main, mode):
        super(UniquificationPass, self).__init__(main)
        self.mode = mode
        self.seen = {}
        self.original_names = {}

    def __call__(self, definition):
        name = definition.name
        key = hash(definition)
        insert = False
        if name not in self.seen:
            self.seen[name] = {}
            insert = True
        elif self.mode is UniquificationMode.UNIQUIFY:
            seen = self.seen[name]
            if key in seen:
                new_name = seen[key].name
                insert = False
            else:
                idx = len(seen)
                new_name = name + "_unq" + str(idx)
                insert = True
            self.original_names[definition] = name
            type(definition).rename(definition, new_name)
        else:
            insert = True
        if insert:
            self.seen[name][key] = definition

    def _run(self, definition):
        if not isdefinition(definition):
            return

        for instance in definition.instances:
            instancedefinition = type(instance)
            if isdefinition(instancedefinition):
                self._run( instancedefinition )

        self(definition)

    def run(self):
        super(UniquificationPass, self).run()
        duplicated = []
        for name, definitions in self.seen.items():
            if len(definitions) > 1:
                duplicated.append((name, definitions))
        UniquificationPass.handle(duplicated, self.mode)

    @staticmethod
    def handle(duplicated, mode):
        if len(duplicated):
            msg = f"Multiple definitions: {[name for name, _ in duplicated]}"
            if mode is UniquificationMode.ERROR:
                error(msg)
                raise MultipleDefinitionException([name for name, _ in duplicated])
            elif mode is UniquificationMode.WARN:
                warning(msg)


# This pass runs uniquification according to @mode and returns a dictionary
# mapping any renamed circuits to their original names. If @mode is ERROR or
# WARN the returned dictionary should be empty.
def uniquification_pass(circuit, mode):
    pass_ = UniquificationPass(circuit, mode)
    pass_.run()
    return pass_.original_names



from __future__ import division
from itertools import chain
from collections import OrderedDict
from .ref import AnonRef, InstRef, DefnRef
from .t import Type, Kind, In, Out, Flip
from .bit import BitKind, MakeBit
from .port import INPUT, OUTPUT, INOUT
#from .bit import *
from .clock import ClockType, ClockTypes
from .array import ArrayType
from .tuple import TupleType
from .compatibility import IntegerTypes, StringTypes

__all__  = ['DeclareInterface']
__all__ += ['Interface']
__all__ += ['InterfaceKind']

# flatten an iterable of iterables to list
def flatten(l):
    return list(chain(*l))

#
# parse argument declaration of the form
#
#  (name0, type0, name1, type1, ..., namen, typen)
#
def parse(decl):
    #print(decl)
    n = len(decl)
    assert n % 2 == 0

    names = []
    ports = []
    for i in range(0,n,2):
        name = decl[i]   # name
        if not name:
            name = i//2
        port = decl[i+1] # type

        assert isinstance(port, Kind) or isinstance(port, Type)


        names.append(name)
        ports.append(port)

    return names, ports

#
# Abstract Base Class for an Interface
#
class _Interface(Type):

    def __str__(self):
        return str(type(self))

    def __repr__(self):
        s = ""
        for name, input in self.ports.items():
            if input.isinput():
                output = input.value()
                if isinstance(output, ArrayType) \
                  or isinstance(output, TupleType):
                    if not output.iswhole(output.ts):
                        for i in range(len(input)):
                            iname = repr( input[i] )
                            oname = repr( output[i] )
                            s += 'wire({}, {})\n'.format(oname, iname)
                        continue
                iname = repr( input )
                oname = repr( output )
                s += 'wire({}, {})\n'.format(oname, iname)
        return s

    @classmethod
    def items(cls):
        return cls.ports.items()

    def __iter__(self):
        return iter(self.ports)

    def __len__(self):
        return len(self.ports.keys())

    def __getitem__(self, key):
        if isinstance(key, int):
            if isinstance(key,slice):
                return array([self[i] for i in range(*key.indices(len(self)))])
            else:
                n = len(self)
                assert -n < key and key < n, "key: %d, self.N: %d" %(key,len(self))
                return self.arguments()[key]
        else:
            assert isinstance(key, str)
            return self.ports[key]

    # return all the argument ports
    def arguments(self):
        return [port  for name, port in self.ports.items()]

    # return all the argument input ports
    def inputs(self, include_clocks=False):
        return [port for name, port in self.ports.items() \
                    if port.isinput() and (not isinstance(port, ClockTypes) or include_clocks) ]
#                                    name not in ['SET', 'CIN']]

    # return all the argument output ports
    def outputs(self):
        return [port for name, port in self.ports.items() if port.isoutput()]


    # return all the arguments as name, port
    def args(self):
        return flatten([name, port] for name, port in self.ports.items())

    # return all the arguments as name, flip(port)
    #   same as the declaration
    def decl(self):
        return flatten([name, type(port).flip()] \
                             for name, port in self.ports.items()  )


    # return all the input arguments as name, port
    def inputargs(self):
        return flatten( \
                [name, port] for name, port in self.ports.items() \
                    if port.isinput() and not isinstance(port, ClockTypes) )
#                                    name not in ['SET', 'CIN']] )

    # return all the output arguments as name, port
    def outputargs(self):
        return flatten( [name, port] for name, port in self.ports.items() \
                            if port.isoutput() )

    # return all the clock arguments as name, port
    def clockargs(self):
        return flatten( [name, port] for name, port in self.ports.items() \
                            if isinstance(port, ClockTypes)  )
#                                        or name in ['SET'] ] )

    # return all the clock argument names
    def clockargnames(self):
        return [name for name, port in self.ports.items() \
                    if isinstance(port, ClockTypes) ]
#                                or name in ['SET'] ]


    # return True if this interface has a Clock
    def isclocked(self):
        for name, port in self.ports.items():
            if isinstance(port, ClockType):
                return True
        return False

#
# Interface class
#
# This function assumes the port instances are provided
#
#  e.g. Interface('I0', In(Bit)(), 'I1', In(Bit)(), 'O', Out(Bit)())
#
class Interface(_Interface):
    def __init__(self, decl, renamed_ports={}):

        names, ports = parse(decl)

        # setup ports
        args = OrderedDict()

        for name, port in zip(names, ports):

            if isinstance(name, IntegerTypes):
                name = str(name) # convert integer to str, e.g. 0 to "0"

            if name in renamed_ports:
                raise NotImplementedError()

            args[name] = port

        self.ports = args

    def __str__(self):
        return f'Interface({", ".join(f"{k}: {v}" for k, v in self.ports.items())})'


#
# _DeclareInterface class
#
# First, an Interface is declared
#
#  Interface = DeclareInterface('I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit))
#
# Then, the interface is instanced
#
#  interface = Interface()
#
class _DeclareInterface(_Interface):
    def __init__(self, renamed_ports={}, inst=None, defn=None):

        # parse the class Interface declaration
        names, ports = parse(self.Decl)

        args = OrderedDict()

        for name, port in zip(names, ports):
            if   inst: ref = InstRef(inst, name)
            elif defn: ref = DefnRef(defn, name)
            else:      ref = AnonRef(name)

            if hasattr(port, "origPortName"):
                ref.name = port.origPortName
            if name in renamed_ports:
                ref.name = renamed_ports[name]

            if defn:
               port = port.flip()

            args[name] = port(name=ref)

        self.ports = args

class InterfaceKind(Kind):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ports = []
        key = None
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                key = arg
            else:
                ports.append((key, arg))
        cls.ports = OrderedDict(ports)

    def items(cls):
        return cls.ports.items()

    def __iter__(cls):
        return iter(cls.ports)

    def __str__(cls):
        args = []
        for i, arg in enumerate(cls.Decl):
            if i % 2 == 0:
                args.append('"{}"'.format(arg))
            else:
                args.append(str(arg))
        return ', '.join(args)

    def __eq__(cls, rhs):
        if not isinstance(rhs, InterfaceKind): return False

        if cls.Decl != rhs.Decl: return False
        return True

    __ne__=Kind.__ne__
    __hash__=Kind.__hash__


#
# Interface factory
#
def DeclareInterface(*decl, **kwargs):
    name = '%s(%s)' % ('Interface', ', '.join([str(a) for a in decl]))
    #print('DeclareInterface', name)
    dct = dict(Decl=decl, **kwargs)
    return InterfaceKind(name, (_DeclareInterface,), dct)




from collections import Sequence, Mapping, OrderedDict
from .ref import AnonRef, TupleRef
from .t import Type, Kind
from .compatibility import IntegerTypes
from .bit import BitOut, VCC, GND
from .debug import debug_wire, get_callee_frame_info
from .port import report_wiring_error

__all__  = ['TupleType', 'TupleKind', 'Tuple']

#
# Create an Tuple
#
#  Tuple()
#  - creates a new tuple value
#  Tuple(v0, v1, ..., vn)
#  - creates a new tuple value where each field equals vi
#
class TupleType(Type):
    def __init__(self, *largs, **kwargs):

        Type.__init__(self, **kwargs) # name=

        self.ts = []
        if len(largs) > 0:
            assert len(largs) == self.N
            for i in range(self.N):
                k = str(self.Ks[i])
                t = largs[i]
                if isinstance(t, IntegerTypes):
                    t = VCC if t else GND
                assert type(t) == self.Ts[i]
                self.ts.append(t)
                setattr(self, k, t)
        else:
            for i in range(self.N):
                k = str(self.Ks[i])
                T = self.Ts[i]
                t = T(name=TupleRef(self,k))
                self.ts.append(t)
                setattr(self, k, t)

    __hash__ = Type.__hash__

    def __eq__(self, rhs):
        if not isinstance(rhs, TupleType): return False
        return self.ts == rhs.ts

    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        kts = ['{}={}'.format(k, v) for k, v in zip(self.Ks, ts)]
        return 'tuple(dict({})'.format(', '.join(kts))

    def __getitem__(self, key):
        if key in self.Ks:
            key = self.Ks.index(key)
        return self.ts[key]

    def __len__(self):
        return self.N

    def __call__(self, o):
        return self.wire(o, get_callee_frame_info())

    @classmethod
    def isoriented(cls, direction):
        for T in cls.Ts:
            if not T.isoriented(direction):
                return False
        return True


    @debug_wire
    def wire(i, o, debug_info):
        # print('Tuple.wire(', o, ', ', i, ')')

        if not isinstance(o, TupleType):
            report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}) to {i.debug_name} (type={type(i)}) because {o.debug_name} is not a Tuple', debug_info)  # noqa
            return

        if i.Ks != o.Ks:
            report_wiring_error(f'Cannot wire {o.debug_name} (type={type(o)}, keys={i.Ks}) to {i.debug_name} (type={type(i)}, keys={o.Ks}) because the tuples do not have the same keys', debug_info)  # noqa
            return

        #if i.Ts != o.Ts:
        #    print('Wiring error: Tuple elements must have the same type')
        #    return

        for i_elem, o_elem in zip(i, o):
            if o_elem.isinput():
                o_elem.wire(i_elem, debug_info)
            else:
                i_elem.wire(o_elem, debug_info)

    def driven(self):
        for t in self.ts:
            if not t.driven():
                return False
        return True

    def wired(self):
        for t in self.ts:
            if not t.wired():
                return False
        return True

    # test whether the values refer a whole tuple
    def iswhole(self, ts):

        for i in range(len(ts)):
            if ts[i].anon():
                #print('not an inst or defn')
                return False

        for i in range(len(ts)):
            # elements must be an tuple reference
            if not isinstance(ts[i].name, TupleRef):
                #print('not an tuple ref')
                return False

        for i in range(1,len(ts)):
            # elements must refer to the same tuple
            if ts[i].name.tuple is not ts[i-1].name.tuple:
                return False

        for i in range(len(ts)):
            # elements should be numbered consecutively
            if ts[i].name.index != self.Ks[i]:
                return False

        return True


    def trace(self):
        ts = [t.trace() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == self.N and self.iswhole(ts):
            return ts[0].name.tuple

        return tuple_(dict(zip(self.Ks,ts)))

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if len(ts) == self.N and self.iswhole(ts):
            return ts[0].name.tuple

        return tuple_(dict(zip(self.Ks,ts)))

    def flatten(self):
        return sum([t.flatten() for t in self.ts], [])

    def const(self):
        for t in self.ts:
            if not t.const():
                return False

        return True

class TupleKind(Kind):
    def __init__(cls, name, bases, dct):
        super(TupleKind, cls).__init__(name, bases, dct)
        cls.N = len(cls.Ks)

    def __str__(cls):
        args = []
        for i in range(cls.N):
             args.append("%s=%s" % (cls.Ks[i], str(cls.Ts[i])))
        return "Tuple(%s)" % ",".join(args)

    def __eq__(cls, rhs):
        if not isinstance(rhs, TupleKind): return False

        if cls.Ks != rhs.Ks: return False
        if cls.Ts != rhs.Ts: return False

        return True

    __ne__ = Kind.__ne__
    __hash__ = Kind.__hash__

    def __len__(cls):
        return cls.N

    def __getitem__(cls, key):
        if key in cls.Ks:
            key = cls.Ks.index(key)
        return cls.Ts[key]

    def size(cls):
        n = 0
        for T in cls.Ts:
            n += T.size()
        return n

    def qualify(cls, direction):
        if cls.isoriented(direction):
            return cls
        return Tuple(OrderedDict(zip(cls.Ks, [T.qualify(direction) for T in cls.Ts])))

    def flip(cls):
        return Tuple(OrderedDict(zip(cls.Ks, [T.flip() for T in cls.Ts])))


#
# Tuple(Mapping)
# Tuple(Sequence)
#
# *largs with largs being comprised of magma types
#    Tuple(T0, T1, ..., Tn)
#
# **kwargs - only called if largs is empty
#    Tuple(x=Bit, y=Bit) -> Tuple(**kwargs)
#
def Tuple(*largs, **kwargs):
    if largs:
        if isinstance(largs[0], Kind):
            Ks = range(len(largs))
            Ts = largs
        else:
            largs = largs[0]
            if isinstance(largs, Sequence):
                Ks = range(len(largs))
                Ts = largs
            elif isinstance(largs, Mapping):
                Ks = list(largs.keys())
                Ts = list(largs.values())
            else:
                assert False
    else:
        Ks = list(kwargs.keys())
        Ts = list(kwargs.values())

    # check types and promote integers
    for i in range(len(Ts)):
        T = Ts[i]
        if T in IntegerTypes:
            T = BitOut
            Ts[i] = T
        if not isinstance(T, Kind):
            raise ValueError(f'Tuples must contain magma types - got {T}')

    name = 'Tuple(%s)' % ", ".join([f"{k}: {t}" for k, t in zip(Ks, Ts)])
    return TupleKind(name, (TupleType,), dict(Ks=Ks, Ts=Ts))

from .bitutils import int2seq
from .array import ArrayType, Array
from .bit import _BitKind, _BitType, Bit, BitKind, BitType, VCC, GND

#
# convert value to a tuple
#   *value = tuple from positional arguments
#   **kwargs = tuple from keyword arguments
#
def tuple_(value, n=None):
    if isinstance(value, TupleType):
        return value

    if not isinstance(value, (_BitType, ArrayType, IntegerTypes, Sequence, Mapping)):
        raise ValueError(
            "bit can only be used on a Bit, an Array, or an int; not {}".format(type(value)))

    decl = OrderedDict()
    args = []

    if isinstance(value, IntegerTypes):
        if n is None:
            n = max(value.bit_length(),1)
        value = int2seq(value, n)
    elif isinstance(value, _BitType):
        value = [value]
    elif isinstance(value, ArrayType):
        value = [value[i] for i in range(len(value))]

    if isinstance(value, Sequence):
        ts = list(value)
        for i in range(len(ts)):
            args.append(ts[i])
            decl[i] = type(ts[i])
    elif isinstance(value, Mapping):
        for k, v in value.items():
            args.append(v)
            decl[k] = type(v)

    return Tuple(decl)(*args)


def namedtuple(**kwargs):
    return tuple_(kwargs)



from __future__ import absolute_import
from __future__ import print_function
from collections import namedtuple

from mako.template import Template
from pyverilog.vparser.parser import VerilogParser, Node, Input, Output, ModuleDef, Ioport, Port, Decl
import pyverilog.vparser.parser as parser
from pyverilog.dataflow.visit import NodeVisitor
import pyverilog.vparser.ast as pyverilog_ast

from .t import In, Out, InOut
from .bit import Bit, _BitKind
from .bits import Bits, BitsKind
from .circuit import DeclareCircuit, DefineCircuit, EndDefine
import logging

logger = logging.getLogger('magma').getChild('from_verilog')

__all__  = ['DeclareFromVerilog']
__all__ += ['DeclareFromVerilogFile']
__all__ += ['DeclareFromTemplatedVerilog']
__all__ += ['DeclareFromTemplatedVerilogFile']
__all__ += ['DefineFromVerilog']
__all__ += ['DefineFromVerilogFile']
__all__ += ['DefineFromTemplatedVerilog']
__all__ += ['DefineFromTemplatedVerilogFile']


class ModuleVisitor(NodeVisitor):
    def __init__(self):
        self.nodes = []

    def visit_ModuleDef(self, node):
        self.nodes.append(node)
        return node

def convert(input_type, target_type):
    if isinstance(input_type, _BitKind) and \
       isinstance(target_type, _BitKind) and \
       input_type.direction == target_type.direction:
        return target_type
    if isinstance(input_type, BitsKind) and \
       isinstance(target_type, BitsKind) and \
       input_type.N == target_type.N and \
       input_type.T.direction == target_type.T.direction:
        return target_type
    raise NotImplementedError(f"Conversion between {input_type} and "
                              f"{target_type} not supported")

def get_value(v):
    if isinstance(v, pyverilog_ast.IntConst):
        return int(v.value)
    if isinstance(v, pyverilog_ast.Minus):
        return get_value(v.left) - get_value(v.right)
    else:
        raise NotImplementedError(type(v))

def get_type(io, type_map):
    if isinstance(io, Input):
        direction = In
    elif isinstance(io, Output):
        direction = Out
    else:
        direction = InOut

    if io.width is None:
        type_ = Bit
    else:
        msb = get_value(io.width.msb)
        lsb = get_value(io.width.lsb)
        type_ = Bits(msb-lsb+1)

    type_ = direction(type_)

    if io.name in type_map:
        type_ = convert(type_, type_map[io.name])

    return type_


def ParseVerilogModule(node, type_map):
    args = []
    ports = []
    for port in node.portlist.ports:
        if isinstance(port, Ioport):
            io = port.first
            args.append(io.name)
            args.append(get_type(io, type_map))
        elif isinstance(port, Port):
            ports.append(port.name)
        else:
            raise NotImplementedError(type(port))

    if ports:
        assert not args, "Can we have mixed declared and undeclared types in a Verilog module?"
        for port in ports:
            found = False
            for child in node.children():
                if isinstance(child, Decl):
                    for sub_child in child.children():
                        if isinstance(sub_child, (parser.Input, parser.Output, parser.Inout)) and \
                                sub_child.name == port:
                            args.append(sub_child.name)
                            args.append(get_type(sub_child, type_map))
                            found = True
                            break
                if found:
                    break
            else:
                raise Exception(f"Could not find type declaration for port {port}")

    return node.name, args

def FromVerilog(source, func, type_map, target_modules=None):
    parser = VerilogParser()

    ast = parser.parse(source)
    #ast.show()

    v = ModuleVisitor()
    v.visit(ast)

    if func == DefineCircuit:
        # Only allow a single verilog module unless we're only defining one
        # circuit (only one module in target_modules), otherwise, they would all
        # use the same source, so if they are compiled together, there will be
        # multiple definitions of the same verilog module.
        assert len(v.nodes) == 1 or (target_modules and len(target_modules) == 1)
    modules = []
    for node in v.nodes:
        if target_modules is not None and node.name not in target_modules:
            continue
        try:
            name, args = ParseVerilogModule(node, type_map)
            circuit = func(name, *args)
            if func == DefineCircuit:
                # inline source
                circuit.verilogFile = source
                EndDefine()
            circuit.verilog_source = source
            modules.append(circuit)
        except Exception as e:
            logger.warning(f"Could not parse module {node.name} ({e}), "
                           f"skipping")
    if not modules:
        logger.warning(f"Did not import any modules from verilog, either could "
                       f"not parse or could not find any of the target_modules "
                       f"({target_modules})")
    return modules

def FromVerilogFile(file, func, type_map, target_modules=None):
    if file is None:
        return None
    verilog = open(file).read()
    result = FromVerilog(verilog, func, type_map, target_modules)
    # Store the original verilog file name, currently used by m.compile to
    # generate a .sv when compiling a circuit that was defined from a verilog
    # file
    for item in result:
        item.verilog_file_name = file
    return result

def FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs):
    verilog = Template(templatedverilog).render(**kwargs)
    return FromVerilog(verilog, func, type_map)

def FromTemplatedVerilogFile(file, func, type_map, **kwargs):
    if file is None:
        return None
    templatedverilog = open(file).read()
    return FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs)


def DeclareFromVerilog(source, type_map={}):
    return FromVerilog(source, DeclareCircuit, type_map)

def DeclareFromVerilogFile(file, target_modules=None, type_map={}):
    return FromVerilogFile(file, DeclareCircuit, type_map, target_modules)

def DeclareFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DeclareCircuit, type_map, **kwargs)

def DeclareFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DeclareCircuit, type_map, **kwargs)


def DefineFromVerilog(source, type_map={}, target_modules=None):
    return FromVerilog(source, DefineCircuit, type_map, target_modules)

def DefineFromVerilogFile(file, target_modules=None, type_map={}):
    return FromVerilogFile(file, DefineCircuit, type_map, target_modules)

def DefineFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DefineCircuit, type_map, **kwargs)

def DefineFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DefineCircuit, type_map, **kwargs)




import magma as m


def BitOrBits(width):
    if width is None:
        return m.Bit
    if not isinstance(width, int):
        raise ValueError(f"Expected width to be None or int, got {width}")
    return m.Bits(width)



from .compatibility import IntegerTypes
from .ref import AnonRef
from .bit import Bit, VCC, GND
from .array import ArrayType, ArrayKind
from .debug import debug_wire

from bit_vector import BitVector, SIntVector

__all__ = ['Bits', 'BitsType', 'BitsKind']
__all__ += ['UInt', 'UIntType', 'UIntKind']
__all__ += ['SInt', 'SIntType', 'SIntKind']


class BitsType(ArrayType):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'bits([{}])'.format(', '.join(ts))

    def bits(self):
        if not self.const():
            raise Exception("Not a constant")

        def convert(x):
            if x is VCC:
                return True
            assert x is GND
            return False
        return [convert(x) for x in self.ts]

    def __int__(self):
        if not self.const():
            raise Exception("Can't call __int__ on a non-constant")
        return BitVector(self.bits()).as_int()

    @debug_wire
    def wire(i, o, debug_info):
        if isinstance(o, IntegerTypes):
            if o.bit_length() > len(i):
                raise ValueError(
                    f"Cannot convert integer {o} (bit_length={o.bit_length()}) to Bits({len(i)})")
            from .conversions import bits
            o = bits(o, len(i))
        super().wire(o, debug_info)


class BitsKind(ArrayKind):
    def __str__(cls):
        if cls.isinput():
            return "In(Bits({}))".format(cls.N)
        if cls.isoutput():
            return "Out(Bits({}))".format(cls.N)
        return "Bits({})".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return Bits(cls.N, cls.T.qualify(direction))

    def flip(cls):
        return Bits(cls.N, cls.T.flip())


def Bits(N, T=None):
    if T is None:
        T = Bit
    assert isinstance(N, IntegerTypes)
    name = 'Bits({})'.format(N)
    return BitsKind(name, (BitsType,), dict(N=N, T=T))


class UIntType(BitsType):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'uint([{}])'.format(', '.join(ts))


class UIntKind(BitsKind):
    def __str__(cls):
        if cls.isinput():
            return "In(UInt({}))".format(cls.N)
        if cls.isoutput():
            return "Out(UInt({}))".format(cls.N)
        return "UInt({})".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return UInt(cls.N, cls.T.qualify(direction))

    def flip(cls):
        return UInt(cls.N, cls.T.flip())


def UInt(N, T=None):
    if T is None:
        T = Bit
    assert isinstance(N, IntegerTypes)
    name = 'UInt({})'.format(N)
    return UIntKind(name, (UIntType,), dict(N=N, T=T))


class SIntType(BitsType):
    def __repr__(self):
        if not isinstance(self.name, AnonRef):
            return repr(self.name)
        ts = [repr(t) for t in self.ts]
        return 'sint([{}])'.format(', '.join(ts))

    def __int__(self):
        if not self.const():
            raise Exception("Can't call __int__ on a non-constant")
        return SIntVector(self.bits()).as_sint()


class SIntKind(BitsKind):
    def __str__(cls):
        if cls.isinput():
            return "In(SInt({}))".format(cls.N)
        if cls.isoutput():
            return "Out(SInt({}))".format(cls.N)
        return "SInt({})".format(cls.N)

    def qualify(cls, direction):
        if cls.T.isoriented(direction):
            return cls
        return SInt(cls.N, cls.T.qualify(direction))

    def flip(cls):
        return SInt(cls.N, cls.T.flip())


def SInt(N, T=None):
    if T is None:
        T = Bit
    assert isinstance(N, IntegerTypes)
    name = 'SInt({})'.format(N)
    return SIntKind(name, (SIntType,), dict(N=N, T=T))



from .logging import error, warning, get_source_line
from .backend.util import make_relative
from .ref import DefnRef, InstRef

__all__  = ['INPUT', 'OUTPUT', 'INOUT']
__all__ += ['flip']
__all__ += ['Port']

INPUT = 'input'
OUTPUT = 'output'
INOUT = 'inout'


def report_wiring_error(message, debug_info):
    error(f"\033[1m{make_relative(debug_info[0])}:{debug_info[1]}: {message}",
          include_wire_traceback=True)
    try:
        error(get_source_line(debug_info[0], debug_info[1]))
    except FileNotFoundError:
        error(f"    Could not find file {debug_info[0]}")


def report_wiring_warning(message, debug_info):
    # TODO: Include wire traceback support
    warning(f"\033[1m{make_relative(debug_info[0])}:{debug_info[1]}: {message}")
    try:
        warning(get_source_line(debug_info[0], debug_info[1]))
    except FileNotFoundError:
        warning(f"    Could not find file {debug_info[0]}")


def flip(direction):
    assert direction in [INPUT, OUTPUT, INOUT]
    if   direction == INPUT:  return OUTPUT
    elif direction == OUTPUT: return INPUT
    elif direction == INOUT:  return INOUT

def mergewires(new, old, debug_info):
    oldinputs = set(old.inputs)
    newinputs = set(new.inputs)
    oldoutputs = set(old.outputs)
    newoutputs = set(new.outputs)

    for i in oldinputs - newinputs:
        new.inputs.append(i)
        i.wires = new

    for o in oldoutputs - newoutputs:
        if len(new.outputs) > 0:
            outputs = [o.bit.debug_name for o in new.outputs]
            report_wiring_error(f"Connecting more than one output ({outputs}) to an input `{i.bit.debug_name}`", debug_info)  # noqa
        new.outputs.append(o)
        o.wires = new


#
# A Wire has a list of input and output Ports.
#
class Wire:
    def __init__(self):
        self.inputs = []
        self.outputs = []

    def connect( self, o, i , debug_info):

        # anon Ports are added to the input or output list of this wire
        #
        # connecting to a non-anonymous port to an anonymous port
        # add the non-anonymous port to the wire associated with the
        # anonymous port

        #print(str(o), o.anon(), o.bit.isinput(), o.bit.isoutput())
        #print(str(i), i.anon(), i.bit.isinput(), i.bit.isoutput())

        if not o.anon():
            #assert o.bit.direction is not None
            if o.bit.isinput():
                report_wiring_error(f"Using `{o.bit.debug_name}` (an input) as an output", debug_info)
                return

            if o not in self.outputs:
                if len(self.outputs) != 0:
                    warn_str = "Adding the output `{}` to the wire `{}` which already has output(s) `[{}]`".format(o.bit.debug_name, i.bit.debug_name, ", ".join(output.bit.debug_name for output in self.outputs))
                    report_wiring_warning(warn_str, debug_info)  # noqa
                #print('adding output', o)
                self.outputs.append(o)

        if not i.anon():
            #assert i.bit.direction is not None
            if i.bit.isoutput():
                report_wiring_error(f"Using `{i.bit.debug_name}` (an output) as an input", debug_info)
                return

            if i not in self.inputs:
                #print('adding input', i)
                self.inputs.append(i)

        # print(o.wires,i.wires,self,self.outputs,self.inputs)

        # always update wires
        o.wires = self
        i.wires = self

    def check(self):
        for o in self.inputs:
            if o.isoutput():
                error("Output in the wire inputs: {}".format(o))

        for o in self.outputs:
            if o.isinput():
                error("Input in the wire outputs: {}".format(o))
                return False

        # check that this wire is only driven by a single output
        if len(self.outputs) > 1:
            error("Multiple outputs on a wire: {}".format(self.outputs))
            return False

        return True

#
# Port implements wiring
#
# Each port is represented by a Bit()
#
class Port:
    def __init__(self, bit):

        self.bit = bit

        self.wires = Wire()

    def __repr__(self):
        return repr(self.bit)

    def __str__(self):
        return str(self.bit)

    def anon(self):
        return self.bit.anon()

    # wire a port to a port
    def wire(i, o, debug_info):
        #if o.bit.direction is None:
        #    o.bit.direction = OUTPUT
        #if i.bit.direction is None:
        #    i.bit.direction = INPUT

        #print("Wiring", o.bit.direction, str(o), "->", i.bit.direction, str(i))

        if i.wires and o.wires and i.wires is not o.wires:
            # print('merging', i.wires.inputs, i.wires.outputs)
            # print('merging', o.wires.inputs, o.wires.outputs)
            w = Wire()
            mergewires(w, i.wires, debug_info)
            mergewires(w, o.wires, debug_info)
            # print('after merge', w.inputs, w.outputs)
        elif o.wires:
            w = o.wires
        elif i.wires:
            w = i.wires
        else:
            w = Wire()

        w.connect(o, i, debug_info)

        #print("after",o,"->",i, w)


    # if the port is an input or inout, return the output
    # if the port is an output, return the first input
    def trace(self):
        if not self.wires:
            return None

        if self in self.wires.inputs:
            if len(self.wires.outputs) < 1:
                # print('Warning:', str(self), 'is not connected to an output')
                return None
            assert len(self.wires.outputs) == 1
            return self.wires.outputs[0]

        if self in self.wires.outputs:
            if len(self.wires.inputs) < 1:
                # print('Warning:', str(self), 'is not connected to an input')
                return None
            assert len(self.wires.inputs) == 1
            return self.wires.inputs[0]

        return None

    # if the port is in the inputs, return the output
    def value(self):
        if not self.wires:
            return None

        if self in self.wires.inputs:
            if len(self.wires.outputs) < 1:
                # print('Warning:', str(self), 'is not connected to an output')
                return None
            #assert len(self.wires.outputs) == 1
            return self.wires.outputs[0]

        return None


    def driven(self):
        return self.value() is not None

    def wired(self):
        return self.trace() is not None




__COMPILE_DIR = 'normal'


def set_compile_dir(target):
    global __COMPILE_DIR
    assert target in ['normal', 'callee_file_dir']
    __COMPILE_DIR = target


def get_compile_dir():
    return __COMPILE_DIR



import inspect
from collections import Sequence
from .port import INPUT, OUTPUT, INOUT
from .compatibility import IntegerTypes
from .t import Type
from .debug import debug_wire
from .logging import info, warning, error
from .port import report_wiring_error

__all__ = ['wire']


@debug_wire
def wire(o, i, debug_info):

    # Wire(o, Circuit)
    if hasattr(i, 'interface'):
        i.wire(o, debug_info)
        return

    # replace output Circuit with its output (should only be 1 output)
    if hasattr(o, 'interface'):
        # if wiring a Circuit to a Port
        # then circuit should have 1 output
        o_orig = o
        o = o.interface.outputs()
        if len(o) != 1:
            report_wiring_error(f'Can only wire circuits with one output. Argument 0 to wire `{o_orig.debug_name}` has outputs {o}', debug_info)  # noqa
            return
        o = o[0]

    # if o is an input
    if not isinstance(o, IntegerTypes) and o.isinput():
        # if i is not an input
        if isinstance(i, IntegerTypes) or not i.isinput():
            # flip i and o
            i, o = o, i

    #if hasattr(i, 'wire'):
    #    error('Wiring Error: The input must have a wire method -  {} to {}'.format(o, i))
    #    return

    # Wire(o, Type)
    i.wire(o, debug_info)

